# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax version of the regularised GW Solver (Peyre et al. 2016)."""
from typing import Any, Dict, Mapping, NamedTuple, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
from typing_extensions import Literal

from ott.geometry import epsilon_scheduler, geometry, low_rank, pointcloud
from ott.initializers.linear import initializers_lr
from ott.initializers.quadratic import initializers as quad_initializers
from ott.math import fixed_point_loop
from ott.problems.linear import linear_problem
from ott.problems.quadratic import quadratic_costs, quadratic_problem
from ott.solvers import was_solver
from ott.solvers.linear import sinkhorn, sinkhorn_lr

__all__ = ["GWOutput", "GromovWasserstein", "gromov_wasserstein"]

LinearOutput = Union[sinkhorn.SinkhornOutput, sinkhorn_lr.LRSinkhornOutput]


class GWOutput(NamedTuple):
  """Holds the output of the Gromov-Wasserstein solver.

  Args:
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    converged: Bool convergence flag for the outer GW iterations.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
    linear_state: State used to solve and store solutions to the local
      linearization of GW.
    geom: The geometry underlying the local linearization.
    old_transport_mass: Holds total mass of transport at previous iteration.
  """

  costs: Optional[jnp.ndarray] = None
  linear_convergence: Optional[jnp.ndarray] = None
  converged: bool = False
  errors: Optional[jnp.ndarray] = None
  linear_state: Optional[LinearOutput] = None
  geom: Optional[geometry.Geometry] = None
  # Intermediate values.
  old_transport_mass: float = 1.0

  def set(self, **kwargs: Any) -> 'GWOutput':
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  @property
  def matrix(self) -> jnp.ndarray:
    """Transport matrix."""
    return self._rescale_factor * self.linear_state.matrix

  def apply(self, inputs: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Apply the transport to an array; axis=1 for its transpose."""
    return self._rescale_factor * self.linear_state.apply(inputs, axis=axis)

  @property
  def reg_gw_cost(self) -> float:
    """Regularized optimal transport cost of the linearization."""
    return self.linear_state.reg_ot_cost

  @property
  def _rescale_factor(self) -> float:
    return jnp.sqrt(
        self.old_transport_mass / self.linear_state.transport_mass()
    )


class GWState(NamedTuple):
  """Holds the state of the Gromov-Wasserstein solver.

  Attributes:
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    linear_state: State used to solve and store solutions to the local
      linearization of GW.
    linear_pb: Local linearization of the quadratic GW problem.
    old_transport_mass: Intermediary value of the mass of the transport matrix.
    keys: Random keys passed to low-rank initializers at every GW iteration
      when not using warm start.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
  """

  costs: jnp.ndarray
  linear_convergence: jnp.ndarray
  linear_state: LinearOutput
  linear_pb: linear_problem.LinearProblem
  old_transport_mass: float
  keys: Optional[jnp.ndarray] = None
  errors: Optional[jnp.ndarray] = None

  def set(self, **kwargs: Any) -> 'GWState':
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  def update(
      self, iteration: int, linear_sol: LinearOutput,
      linear_pb: linear_problem.LinearProblem, store_errors: bool,
      old_transport_mass: float
  ) -> 'GWState':
    costs = self.costs.at[iteration].set(linear_sol.reg_ot_cost)
    errors = None
    if store_errors and self.errors is not None:
      errors = self.errors.at[iteration, :].set(linear_sol.errors)
    linear_convergence = self.linear_convergence.at[iteration].set(
        linear_sol.converged
    )

    return self.set(
        linear_state=linear_sol,
        linear_pb=linear_pb,
        costs=costs,
        linear_convergence=linear_convergence,
        errors=errors,
        old_transport_mass=old_transport_mass
    )


@jax.tree_util.register_pytree_node_class
class GromovWasserstein(was_solver.WassersteinSolver):
  """Gromov-Wasserstein solver.

  Args:
    args: Positional_arguments for
      :class:`~ott.solvers.was_solver.WassersteinSolver`.
    warm_start: Whether to initialize (low-rank) Sinkhorn calls using values
      from the previous iteration. If `None`, warm starts are not used for
      standard Sinkhorn, but used for low-rank Sinkhorn.
    quad_initializer: Quadratic initializer. If the solver is entropic,
      :class:`~ott.initializers.quadratic.initializers.QuadraticInitializer`
      is always used. Otherwise, the quadratic initializer wraps the low-rank
      Sinkhorn initializers:

        - `'random'` - :class:`~ott.initializers.linear.initializers_lr.RandomInitializer`.
        - `'rank2'` - :class:`~ott.initializers.linear.initializers_lr.Rank2Initializer`.
        - `'k-means'` - :class:`~ott.initializers.linear.initializers_lr.KMeansInitializer`.
        - `'generalized-k-means'` - :class:`~ott.initializers.linear.initializers_lr.GeneralizedKMeansInitializer`.

      If `None`, the low-rank initializer will be selected in a problem-specific
      manner:

        - if both :attr:`~ott.problems.quadratic.quadratic_problem.QuadraticProblem.geom_xx`
          and :attr:`~ott.problems.quadratic.quadratic_problem.QuadraticProblem.geom_yy`
          are :class:`~ott.geometry.pointcloud.PointCloud` or :class:`~ott.geometry.low_rank.LRCGeometry`,
          :class:`~ott.initializers.linear.initializers_lr.KMeansInitializer`
          is used.
        - otherwise, use :class:`~ott.initializers.linear.initializers_lr.RandomInitializer`.

    kwargs_init: Keyword arguments when creating the initializer.
    kwargs: Keyword arguments for
      :class:`~ott.solvers.was_solver.WassersteinSolver`.
  """

  def __init__(
      self,
      *args: Any,
      warm_start: Optional[bool] = None,
      quad_initializer: Optional[
          Union[Literal["random", "rank2", "k-means", "generalized-k-means"],
                quad_initializers.BaseQuadraticInitializer]] = None,
      kwargs_init: Optional[Mapping[str, Any]] = None,
      **kwargs: Any
  ):
    super().__init__(*args, **kwargs)
    self._warm_start = warm_start
    self.quad_initializer = quad_initializer
    self.kwargs_init = {} if kwargs_init is None else kwargs_init

  def __call__(
      self,
      prob: quadratic_problem.QuadraticProblem,
      init: Optional[linear_problem.LinearProblem] = None,
      key: Optional[jnp.ndarray] = None,
      **kwargs: Any,
  ) -> GWOutput:
    """Run the Gromov-Wasserstein solver.

    Args:
      prob: Quadratic OT problem.
      init: Initial linearization of the quadratic problem. If `None`, it will
        be computed using the initializer.
      kwargs: Keyword arguments used when calling the initializer.

    Returns:
      The Gromov-Wasserstein output.
    """
    if key is None:
      key = jax.random.PRNGKey(0)
    key1, key2 = jax.random.split(key, 2)

    # consider converting problem first if using low-rank solver
    if self.is_low_rank and prob._is_low_rank_convertible:
      prob = prob.to_low_rank()

    if init is None:
      initializer = self.create_initializer(prob)
      init = initializer(prob, epsilon=self.epsilon, key=key1, **kwargs)

    gromov_fn = jax.jit(iterations) if self.jit else iterations
    out = gromov_fn(self, prob, init, key2)
    # TODO(lpapaxanthos): remove stop_gradient when using backprop
    if self.is_low_rank:
      linearization = prob.update_lr_linearization(
          jax.lax.stop_gradient(out.linear_state)
      )
    else:
      linearization = prob.update_linearization(
          jax.lax.stop_gradient(out.linear_state), self.epsilon,
          jax.lax.stop_gradient(out.old_transport_mass)
      )
    linear_state = out.linear_state.set_cost(linearization, True, True)
    iteration = jnp.sum(out.costs != -1)
    converged = jnp.logical_and(
        iteration < self.max_iterations, jnp.all(out.linear_convergence)
    )
    return out.set(linear_state=linear_state, converged=converged)

  def init_state(
      self,
      prob: quadratic_problem.QuadraticProblem,
      init: linear_problem.LinearProblem,
      key: jnp.ndarray,
  ) -> GWState:
    """Initialize the state of the Gromov-Wasserstein iterations.

    Args:
      prob: Quadratic OT problem.
      init: Initial linearization of the quadratic problem.
      key: Random key for low-rank initializers. Only used when
        :attr:`warm_start` is `False`.

    Returns:
      The initial Gromov-Wasserstein state.
    """
    linear_state = self.linear_ot_solver(init)
    num_iter = self.max_iterations
    transport_mass = prob.init_transport_mass()
    if self.store_inner_errors:
      errors = -jnp.ones((num_iter, self.linear_ot_solver.outer_iterations))
    else:
      errors = None

    return GWState(
        costs=-jnp.ones((num_iter,)),
        linear_convergence=-jnp.ones((num_iter,)),
        linear_state=linear_state,
        linear_pb=init,
        old_transport_mass=transport_mass,
        keys=jax.random.split(key, num_iter),
        errors=errors,
    )

  def output_from_state(self, state: GWState) -> GWOutput:
    """Create an output from a loop state.

    Arguments:
      state: A GWState.

    Returns:
      A GWOutput.
    """
    return GWOutput(
        costs=state.costs,
        linear_convergence=state.linear_convergence,
        errors=state.errors,
        linear_state=state.linear_state,
        geom=state.linear_pb.geom,
        old_transport_mass=state.old_transport_mass
    )

  def create_initializer(
      self, prob: quadratic_problem.QuadraticProblem
  ) -> quad_initializers.BaseQuadraticInitializer:
    """Create quadratic, possibly low-rank initializer.

    Args:
      prob: Quadratic OT problem used to determine the initializer.

    Returns:
      The initializer.
    """
    if isinstance(
        self.quad_initializer, quad_initializers.BaseQuadraticInitializer
    ):
      if self.is_low_rank:
        assert isinstance(
            self.quad_initializer, quad_initializers.LRQuadraticInitializer
        ), f"Expected quadratic initializer to be low rank, " \
           f"found `{type(self.quad_initializer).__name__}`."
        assert self.quad_initializer.rank == self.rank, \
            f"Expected quadratic initializer of rank `{self.rank}`, " \
            f"found `{self.quad_initializer.rank}`."
      return self.quad_initializer

    if self.is_low_rank:
      if self.quad_initializer is None:
        types = (pointcloud.PointCloud, low_rank.LRCGeometry)
        kind = "k-means" if isinstance(prob.geom_xx, types) and isinstance(
            prob.geom_yy, types
        ) else "random"
      else:
        kind = self.quad_initializer
      linear_lr_init = initializers_lr.LRInitializer.from_solver(
          self, kind=kind, **self.kwargs_init
      )
      return quad_initializers.LRQuadraticInitializer(linear_lr_init)

    return quad_initializers.QuadraticInitializer(**self.kwargs_init)

  @property
  def warm_start(self) -> bool:
    """Whether to initialize (low-rank) Sinkhorn using previous solutions."""
    return self.is_low_rank if self._warm_start is None else self._warm_start

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    children, aux_data = super().tree_flatten()
    aux_data["warm_start"] = self._warm_start
    aux_data["quad_initializer"] = self.quad_initializer
    aux_data["kwargs_init"] = self.kwargs_init
    return children, aux_data


def iterations(
    solver: GromovWasserstein,
    prob: quadratic_problem.QuadraticProblem,
    init: linear_problem.LinearProblem,
    key: jnp.ndarray,
) -> GWOutput:
  """Jittable Gromov-Wasserstein outer loop."""

  def cond_fn(
      iteration: int, solver: GromovWasserstein, state: GWState
  ) -> bool:
    return solver._continue(state, iteration)

  def body_fn(
      iteration: int, solver: GromovWasserstein, state: GWState,
      compute_error: bool
  ) -> GWState:
    del compute_error  # always assumed true for the outer loop of GW

    lin_state = state.linear_state
    if solver.is_low_rank:
      key = state.keys[iteration]
      init = (lin_state.q, lin_state.r,
              lin_state.g) if solver.warm_start else (None, None, None)
      linear_pb = prob.update_lr_linearization(state.linear_state)
      out = solver.linear_ot_solver(linear_pb, init=init, key=key)
    else:
      init = (lin_state.f, lin_state.g) if solver.warm_start else (None, None)
      linear_pb = prob.update_linearization(
          lin_state, solver.epsilon, state.old_transport_mass
      )
      out = solver.linear_ot_solver(linear_pb, init=init)

    old_transport_mass = jax.lax.stop_gradient(
        state.linear_state.transport_mass()
    )
    return state.update(
        iteration, out, linear_pb, solver.store_inner_errors, old_transport_mass
    )

  state = fixed_point_loop.fixpoint_iter(
      cond_fn=cond_fn,
      body_fn=body_fn,
      min_iterations=solver.min_iterations,
      max_iterations=solver.max_iterations,
      inner_iterations=1,
      constants=solver,
      state=solver.init_state(prob, init, key=key)
  )

  return solver.output_from_state(state)


def make(
    epsilon: Union[epsilon_scheduler.Epsilon, float] = 1.,
    rank: int = -1,
    max_iterations: int = 50,
    jit: bool = False,
    warm_start: Optional[bool] = None,
    store_inner_errors: bool = False,
    linear_ot_solver_kwargs: Optional[Mapping[str, Any]] = None,
    threshold: float = 1e-2,
    min_iterations: int = 1,
    **kwargs: Any,
) -> GromovWasserstein:
  """Create a GromovWasserstein solver.

  Args:
    epsilon: a regularization parameter or a epsilon_scheduler.Epsilon object.
    rank: integer used to constrain the rank of GW solutions if >0.
    max_iterations: the maximum number of outer iterations for
      Gromov Wasserstein.
    jit: bool, if True, jits the function.
    warm_start: Whether to initialize (low-rank) Sinkhorn calls using values
      from the previous iteration. If `None`, it's enabled when using low-rank.
    store_inner_errors: whether or not to return all the errors of the inner
      Sinkhorn iterations.
    linear_ot_solver_kwargs: Optionally a dictionary containing the keywords
      arguments for the linear OT solver (e.g. sinkhorn)
    threshold: threshold (progress between two iterate costs) used to stop GW.
    min_iterations: see fixed_point_loop.
    kwargs: additional kwargs for epsilon.

  Returns:
    A GromovWasserstein solver.
  """
  if linear_ot_solver_kwargs is None:
    linear_ot_solver_kwargs = {}

  if rank == -1:
    sink = sinkhorn.make(**linear_ot_solver_kwargs)
  elif rank > 0:
    # `rank` and `epsilon` are arguments of the `sinkhorn_lr` solver. As we are
    # passing them to make, we should not pass them in `linear_ot_solver_kwargs`
    # Therefore, the `rank` or `epsilon` passed to `linear_ot_solver_kwargs` are
    # deleted.
    _ = linear_ot_solver_kwargs.pop('rank', None)
    _ = linear_ot_solver_kwargs.pop('epsilon', None)
    sink = sinkhorn_lr.make(
        rank=rank, epsilon=epsilon, **linear_ot_solver_kwargs
    )
  else:
    raise ValueError(f"Invalid value for `rank={rank}`.")

  return GromovWasserstein(
      epsilon=epsilon,
      rank=rank,
      linear_ot_solver=sink,
      threshold=threshold,
      min_iterations=min_iterations,
      max_iterations=max_iterations,
      jit=jit,
      store_inner_errors=store_inner_errors,
      warm_start=warm_start,
      **kwargs
  )


def gromov_wasserstein(
    geom_xx: geometry.Geometry,
    geom_yy: geometry.Geometry,
    geom_xy: Optional[geometry.Geometry] = None,
    fused_penalty: float = 1.0,
    scale_cost: Optional[Union[bool, float, str]] = False,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    loss: Union[Literal['sqeucl', 'kl'], quadratic_costs.GWLoss] = 'sqeucl',
    tau_a: Optional[float] = 1.0,
    tau_b: Optional[float] = 1.0,
    gw_unbalanced_correction: bool = True,
    ranks: Union[int, Tuple[int, ...]] = -1,
    tolerances: Union[float, Tuple[float, ...]] = 1e-2,
    **kwargs: Any,
) -> GWOutput:
  """Solve a Gromov Wasserstein problem.

  Wrapper that instantiates a quadratic problem (possibly with linear term
  if the problem is fused) and calls a solver to output a solution.

  Args:
    geom_xx: a Geometry object for the first view.
    geom_yy: a second Geometry object for the second view.
    geom_xy: a Geometry object representing the linear cost in FGW.
    fused_penalty: multiplier of the linear term in Fused Gromov Wasserstein,
      i.e. loss = quadratic_loss + fused_penalty * linear_loss.
      Ignored if ``geom_xy`` is not specified.
    scale_cost: option to rescale the cost matrices:

      - if `True`, use the default for each geometry.
      - if `False`, keep the original scaling in geometries.
      - if :class:`str`, use a specific method available in
        :class:`~ott.geometry.geometry.Geometry` or
        :class:`~ott.geometry.pointcloud.PointCloud`.
      - if `None`, do not scale the cost matrices.

    a: jnp.ndarray<float>[num_a,] or jnp.ndarray<float>[batch,num_a] weights.
    b: jnp.ndarray<float>[num_b,] or jnp.ndarray<float>[batch,num_b] weights.
    loss: defaults to the square Euclidean distance. Can also pass 'kl'
      to define the GW loss as KL loss.
    tau_a: float between 0 and 1.0, parameter that controls the strength of the
      KL divergence constraint between the weights and marginals of the
      transport for the first view. If set to 1.0, then it is equivalent to a
      hard constraint and if smaller to a softer constraint.
    tau_b: float between 0 and 1.0, parameter that controls the strength of the
      KL divergence constraint between the weights and marginals of the
      transport for the second view. If set to 1.0, then it is equivalent to a
      hard constraint and if smaller to a softer constraint.
    gw_unbalanced_correction: True (default) if the unbalanced version of
      :cite:`sejourne:21` is used, False if tau_a and tau_b
      only affect the inner Sinkhorn loop.
    ranks: Switch to a low rank approximation of all cost matrices, using
      :meth:`~ott.geometry.geometry.Geometry.to_LRCGeometry`, to gain speed.
      This is only relevant if the geometries of interest are *not*
      :class:`~ott.geometry.pointcloud.PointCloud` with `'sqeucl'` cost
      function, in which case they would be low-rank by construction (as long
      as the sizes of these point clouds is larger than dimension).
      If `-1`, geometries are left as they are, and not converted.
      If :class:`tuple`, these 2 or 3 :class:`int` specify the ranks of
      ``geom_xx``, ``geom_yy`` and ``geom_xy``, respectively. If :class:`int`,
      all 3 geometries are converted using that rank.
    tolerances: Tolerances used when converting geometries to low-rank. Used when
      geometries are *not* :class:`~ott.geometry.pointcloud.PointCloud` with
      `'sqeucl'` cost. If :class:`float`, that tolerance is shared across all
      3 geometries.
    kwargs: Keyword arguments for
      :class:`~ott.solvers.quadratic.gromov_wasserstein.GromovWasserstein`.

  Returns:
    A GromovWassersteinState named tuple.
  """
  prob = quadratic_problem.QuadraticProblem(
      geom_xx,
      geom_yy,
      geom_xy=geom_xy,
      fused_penalty=fused_penalty,
      scale_cost=scale_cost,
      a=a,
      b=b,
      loss=loss,
      tau_a=tau_a,
      tau_b=tau_b,
      gw_unbalanced_correction=gw_unbalanced_correction,
      ranks=ranks,
      tolerances=tolerances
  )
  solver = make(**kwargs)
  return solver(prob)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A class to define a scheduler for the entropic regularization epsilon."""
from typing import Any, Optional

import jax
import jax.numpy as jnp

__all__ = ["Epsilon"]


@jax.tree_util.register_pytree_node_class
class Epsilon:
  """Scheduler class for the regularization parameter epsilon.

  An epsilon scheduler outputs a regularization strength, to be used by in a
  Sinkhorn-type algorithm, at any iteration count. That value is either the
  final, targeted regularization, or one that is larger, obtained by
  geometric decay of an initial value that is larger than the intended target.
  Concretely, the value returned by such a scheduler will consider first
  the max between ``target`` and ``init * target * decay ** iteration``.
  If the ``scale_epsilon`` parameter is provided, that value is used to
  multiply the max computed previously by ``scale_epsilon``.

  Args:
    target: the epsilon regularizer that is targeted.
    scale_epsilon: if passed, used to multiply the regularizer, to rescale it.
    init: initial value when using epsilon scheduling, understood as multiple
      of target value. if passed, ``int * decay ** iteration`` will be used
      to rescale target.
    decay: geometric decay factor, smaller than 1.
  """

  # TODO(michalk8): directly use the defaults instead of `None`
  def __init__(
      self,
      target: Optional[float] = None,
      scale_epsilon: Optional[float] = None,
      init: Optional[float] = None,
      decay: Optional[float] = None
  ):
    self._target_init = .01 if target is None else target
    self._scale_epsilon = 1.0 if scale_epsilon is None else scale_epsilon
    self._init = 1.0 if init is None else init
    self._decay = 1.0 if decay is None else decay

  @property
  def target(self) -> float:
    """Return the final regularizer value of scheduler."""
    return self._target_init * self._scale_epsilon

  def at(self, iteration: Optional[int] = 1) -> float:
    """Return (intermediate) regularizer value at a given iteration."""
    if iteration is None:
      return self.target
    # check the decay is smaller than 1.0.
    decay = jnp.where(self._decay < 1.0, self._decay, 1.0)
    # the multiple is either 1.0 or a larger init value that is decayed.
    multiple = jnp.maximum(self._init * (decay ** iteration), 1.0)
    return multiple * self.target

  def done(self, eps: float) -> bool:
    return eps == self.target

  def done_at(self, iteration: Optional[int]) -> bool:
    return self.done(self.at(iteration))

  def tree_flatten(self):
    return (
        self._target_init, self._scale_epsilon, self._init, self._decay
    ), None

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del aux_data
    return cls(*children)

  @classmethod
  def make(cls, *args: Any, **kwargs: Any) -> "Epsilon":
    """Create or return an Epsilon instance."""
    if isinstance(args[0], cls):
      return args[0]
    else:
      return cls(*args, **kwargs)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Fit 2 GMMs to 2 point clouds using likelihood and (approx) W2 distance.

Suppose we have two large point clouds and want to estimate a coupling and a
W2 distance between them. :cite:`delon:20` propose fitting a GMM to each
point cloud while simultaneously minimizing a Wasserstein-like distance
called MW2 between the fitted GMMs. MW2 is an upper bound on W2,
the Wasserstein distance between the GMMs. Here we implement
their algorithm as well as a generalization that allows for reweightings using
generalized, penalized expectation-maximization
(see section 6.2 of :cite:`delon:20`).

As in `fit_gmm.py`, we assume that the observations $X_0$ and $X_1$ from
batches 0 and 1 are generated by GMMs with parameters $\Theta_0$ and $\Theta_1$,
respectively. We will use $\Theta$ to denote the combined parameters
for the two GMMs. We denote the (unobserved) components that gave rise to the
observations $X_i$ as $Z_i$.

Our goal is to maximize a weighted sum of the likelihood of the observations $X$
under the fitted GMMs and a measure of distance, $MW_2$, between the fitted
GMMs. The problem would be a straightforward maximization exercise if we knew
the components $Z$ that generated each observation $X$. Because the $Z$ are
unobserved, however, we use EM:

We start with an initial estimate of $\Theta$, $\Theta^{(t)}$.

* The E-step: We use the current $\Theta^{(t)}$ to estimate the likelihood of
  all possible cluster attributions for each observation $X$.

* The M-step: We form the function $Q(\Theta|\Theta^{(t)})$,
  the log likelihood of our observations averaged over all possible
  assignments. We then obtain an updated parameter estimate, $\Theta^{(t+1)}$,
  by numerically maximizing the sum of $Q$ and our GMM distance penalty.

It can be shown that if we maximize the penalized $Q$ above, this procedure will
increase or leave unchanged the penalized log likelihood for $\Theta$. We
iterate over these two steps until convergence. Note that the resulting
estimate for $\Theta$ may only be a *local* maximum of the penalized
likelihood function.


Sample usage:

# (Note that we usually initialize a pair to a single GMM that we fit to a
# pooled set, then the two GMMs separate as we optimize the pair.)
pair_init = gaussian_mixture_pair.GaussianMixturePair(
    gmm0=gmm0,
    gmm1=gmm1,
    epsilon=1.e-2,
    tau=1.)
fit_model_em_fn = fit_gmm_pair.get_fit_model_em_fn(
    weight_transport=0.1,
    weight_splitting=1.,
    epsilon=pair_init.epsilon,
    jit=True)
pair, loss = fit_model_em_fn(
    pair=pair_init,
    points0=samples_gmm0,
    points1=samples_gmm1,
    point_weights0=None,
    point_weights1=None,
    em_steps=30,
    m_steps=20,
    verbose=True)
"""
# TODO(geoffd): look into refactoring so we jit higher level functions

import functools
import math
from typing import Callable, NamedTuple, Optional, Tuple

import jax
import jax.numpy as jnp
import optax

from ott.tools.gaussian_mixture import fit_gmm, gaussian_mixture, gaussian_mixture_pair

LOG2 = math.log(2)


class Observations(NamedTuple):
  """Weighted observations and their E-step assignment probabilities."""

  points: jnp.ndarray
  point_weights: jnp.ndarray
  assignment_probs: jnp.ndarray


# Model fit


def get_q(
    gmm: gaussian_mixture.GaussianMixture, obs: Observations
) -> jnp.ndarray:
  r"""Get Q(\Theta|\Theta^{(t)}).

  Here Q is the log likelihood for our observations based on the current
  parameter estimates for \Theta and averaged over the current component
  assignment probabilities. See the overview of EM above for more details.

  Args:
    gmm: GMM model parameterized by Theta
    obs: weighted observations with component assignments computed in the E step
      for \Theta^{(t)}

  Returns:
    Q(\Theta|\Theta^{(t)})
  """
  # Q = E_Z log p(X, Z| Theta)
  # = \sum_Z P(Z|X, Theta^(t)) [log p(X, Z | Theta)]
  # Here P(Z|X, theta^(t)) is the set of assignment probabilities
  # we computed in the E step.
  # log p(X, Z| theta) is given by
  log_p_x_z = (
      gmm.conditional_log_prob(obs.points) +  # p(X | Z, theta)
      gmm.log_component_weights()
  )  # p(Z | theta)
  return (
      jnp.sum(
          obs.point_weights *
          jnp.sum(log_p_x_z * obs.assignment_probs, axis=-1),
          axis=0
      ) / jnp.sum(obs.point_weights, axis=0)
  )


# Objective function


@functools.lru_cache
def get_objective_fn(weight_transport: float):
  """Get the total loss function with static parameters in a closure.

  Args:
    weight_transport: weight for the transport penalty

  Returns:
    A function that returns the objective for a GaussianMixturePair.
  """

  def _objective_fn(
      pair: gaussian_mixture_pair.GaussianMixturePair,
      obs0: Observations,
      obs1: Observations,
  ) -> jnp.ndarray:
    """Compute the objective function for a pair of GMMs.

    Args:
      pair: pair of GMMs + coupling for which to evaluate the objective
      obs0: first set of observations
      obs1: second set of observations

    Returns:
      The objective to be minimized in the M-step.
    """
    q0 = get_q(gmm=pair.gmm0, obs=obs0)
    q1 = get_q(gmm=pair.gmm1, obs=obs1)
    cost_matrix = pair.get_cost_matrix()
    sinkhorn_output = pair.get_sinkhorn(cost_matrix=cost_matrix)
    transport_penalty = sinkhorn_output.reg_ot_cost
    return q0 + q1 - weight_transport * transport_penalty

  return _objective_fn


def print_losses(
    iteration: int, weight_transport: float,
    pair: gaussian_mixture_pair.GaussianMixturePair, obs0: Observations,
    obs1: Observations
):
  """Print the loss components for diagnostic purposes."""
  q0 = get_q(gmm=pair.gmm0, obs=obs0)
  q1 = get_q(gmm=pair.gmm1, obs=obs1)
  cost_matrix = pair.get_cost_matrix()
  sinkhorn_output = pair.get_sinkhorn(cost_matrix=cost_matrix)
  transport_penalty = sinkhorn_output.reg_ot_cost
  objective = q0 + q1 - weight_transport * transport_penalty

  print((
      f'{iteration:3d} {q0:.3f} {q1:.3f} '
      f'transport:{transport_penalty:.3f} '
      f'objective:{objective:.3f}'
  ),
        flush=True)


# The E-step for a single GMM


def do_e_step(
    e_step_fn: Callable[[gaussian_mixture.GaussianMixture, jnp.ndarray],
                        jnp.ndarray],
    gmm: gaussian_mixture.GaussianMixture,
    points: jnp.ndarray,
    point_weights: jnp.ndarray,
) -> Observations:
  assignment_probs = e_step_fn(gmm, points)
  return Observations(
      points=points,
      point_weights=point_weights,
      assignment_probs=assignment_probs
  )


# The M-step


def get_m_step_fn(learning_rate: float, objective_fn, jit: bool):
  """Get a function that performs the M-step of the EM algorithm.

  We precompile and precompute a few quantities that we put into a closure.

  Args:
    learning_rate: learning rate to use for the Adam optimizer
    objective_fn: the objective function to maximize
    jit: if True, precompile key methods

  Returns:
    A function that performs the M-step of EM.
  """
  grad_objective_fn = jax.grad(objective_fn, argnums=(0,))
  gmm_m_step_fn = gaussian_mixture.GaussianMixture.from_points_and_assignment_probs
  if jit:
    grad_objective_fn = jax.jit(grad_objective_fn)
    gmm_m_step_fn = jax.jit(gmm_m_step_fn)

  opt_init, opt_update = optax.chain(
      # Set the parameters of Adam. Note the learning_rate is not here.
      optax.scale_by_adam(b1=0.9, b2=0.999, eps=1e-8),
      optax.scale(learning_rate)
  )

  def _m_step_fn(
      pair: gaussian_mixture_pair.GaussianMixturePair,
      obs0: Observations,
      obs1: Observations,
      steps: int,
  ) -> gaussian_mixture_pair.GaussianMixturePair:
    """Perform the M-step on a pair of Gaussian mixtures.

    Args:
      pair: GMM parameters to optimize
      obs0: first set of observations
      obs1: second set of observations
      steps: number of optimization steps to use when maximizing the objective

    Returns:
      A GaussianMixturePair with updated parameters.
    """
    state = opt_init((pair,))

    for _ in range(steps):
      grad_objective = grad_objective_fn(pair, obs0, obs1)
      updates, state = opt_update(grad_objective, state, (pair,))
      (pair,) = optax.apply_updates((pair,), updates)
      for j, gmm in enumerate((pair.gmm0, pair.gmm1)):
        if gmm.has_nans():
          raise ValueError(f'NaN in gmm{j}')
    return pair

  return _m_step_fn


def get_fit_model_em_fn(
    weight_transport: float,
    learning_rate: float = 0.001,
    jit: bool = True,
):
  """Get a function that performs penalized EM.

  We precompile and precompute a few quantities that we put into a closure.

  Args:
    weight_transport: weight for the transportation loss in the total loss
    learning_rate: learning rate to use for the Adam optimizer
    jit: if True, precompile key methods

  Returns:
    A function that performs generalized, penalized EM.
  """
  objective_fn = get_objective_fn(weight_transport=weight_transport)
  e_step_fn = fit_gmm.get_assignment_probs

  if jit:
    objective_fn = jax.jit(objective_fn)
    e_step_fn = jax.jit(e_step_fn)

  m_step_fn = get_m_step_fn(
      learning_rate=learning_rate, objective_fn=objective_fn, jit=jit
  )

  def _fit_model_em(
      pair: gaussian_mixture_pair.GaussianMixturePair,
      points0: jnp.ndarray,
      points1: jnp.ndarray,
      point_weights0: Optional[jnp.ndarray],
      point_weights1: Optional[jnp.ndarray],
      em_steps: int,
      m_steps: int = 50,
      verbose: bool = False,
  ) -> Tuple[gaussian_mixture_pair.GaussianMixturePair, float]:
    """Optimize a GaussianMixturePair using penalized EM.

    Args:
      pair: GaussianMixturePair to optimize
      points0: observations associated with pair.gmm0
      points1: observations associated with pair.gmm1
      point_weights0: weights for points0
      point_weights1: weights for points1
      em_steps: number of EM steps to perform
      m_steps: number of gradient descent steps to perform in the M-step
      verbose: if True, print status messages

    Returns:
      An updated GaussianMixturePair and the final loss.
    """
    if point_weights0 is None:
      point_weights0 = jnp.ones(points0.shape[0])
    if point_weights1 is None:
      point_weights1 = jnp.ones(points1.shape[0])

    if pair.lock_gmm1:
      obs1 = do_e_step(
          e_step_fn=e_step_fn,
          gmm=pair.gmm1,
          points=points1,
          point_weights=point_weights1
      )

    for i in range(em_steps):
      # E-step
      obs0 = do_e_step(
          e_step_fn=e_step_fn,
          gmm=pair.gmm0,
          points=points0,
          point_weights=point_weights0
      )
      if not pair.lock_gmm1:
        obs1 = do_e_step(
            e_step_fn=e_step_fn,
            gmm=pair.gmm1,
            points=points1,
            point_weights=point_weights1
        )

      # print current losses
      if verbose:
        print_losses(
            iteration=i,
            weight_transport=weight_transport,
            pair=pair,
            obs0=obs0,
            obs1=obs1
        )

      # the M-step
      pair = m_step_fn(pair=pair, obs0=obs0, obs1=obs1, steps=m_steps)

    # final E-step before computing the loss
    obs0 = do_e_step(
        e_step_fn=e_step_fn,
        gmm=pair.gmm0,
        points=points0,
        point_weights=point_weights0
    )
    if not pair.lock_gmm1:
      obs1 = do_e_step(
          e_step_fn=e_step_fn,
          gmm=pair.gmm1,
          points=points1,
          point_weights=point_weights1
      )

    loss = objective_fn(pair=pair, obs0=obs0, obs1=obs1)
    return pair, loss

  return _fit_model_em



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions entering the implicit differentiation of Sinkhorn."""

from typing import TYPE_CHECKING, Callable, Optional, Tuple

import jax
import jax.numpy as jnp

from ott import utils
from ott.math import unbalanced_functions

if TYPE_CHECKING:
  from ott.problems.linear import linear_problem

__all__ = ["ImplicitDiff"]


@utils.register_pytree_node
class ImplicitDiff:
  """Implicit differentiation of Sinkhorn algorithm.

  Args:
    solver_fun: Callable, should return (solution, ...)
    ridge_kernel: promotes zero-sum solutions. only used if tau_a = tau_b = 1.0
    ridge_identity: handles rank deficient transport matrices (this happens
      typically when rows/cols in cost/kernel matrices are collinear, or,
      equivalently when two points from either measure are close).
    symmetric: flag used to figure out whether the linear system solved in the
      implicit function theorem is symmetric or not. This happens when either
      ``a == b`` or the precondition_fun is the identity. False by default, and,
      at the moment, needs to be set manually by the user in the more favorable
      case where the system is guaranteed to be symmetric.
    precondition_fun: TODO(marcocuturi)
  """

  solver_fun: Callable[[jnp.ndarray, jnp.ndarray],
                       Tuple[jnp.ndarray, ...]] = jax.scipy.sparse.linalg.cg
  ridge_kernel: float = 0.0
  ridge_identity: float = 0.0
  symmetric: bool = False
  precondition_fun: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None

  def solve(
      self, gr: Tuple[jnp.ndarray,
                      jnp.ndarray], ot_prob: "linear_problem.LinearProblem",
      f: jnp.ndarray, g: jnp.ndarray, lse_mode: bool
  ) -> jnp.ndarray:
    r"""Apply minus inverse of [hessian ``reg_ot_cost`` w.r.t. ``f``, ``g``].

    This function is used to carry out implicit differentiation of ``sinkhorn``
    outputs, notably optimal potentials ``f`` and ``g``. That differentiation
    requires solving a linear system, using (and inverting) the Jacobian of
    (preconditioned) first-order conditions w.r.t. the reg-OT problem.

    Given a ``precondition_fun``, written here for short as :math:`h`,
    the first order conditions for the dual energy

    .. math::

      E(K, \epsilon, a, b, f, g) :=- <a,\phi_a^{*}(-f)> + <b,
      \phi_b^{*}(-g)> - \langle\exp^{f/\epsilon}, K \exp^{g/\epsilon}>

    form the basis of the Sinkhorn algorithm. To differentiate optimal solutions
    to that problem, we exploit the fact that :math:`h(\nabla E = 0)` and
    differentiate that identity to recover variations (Jacobians) of optimal
    solutions :math:`f^\star, g^\star$` as a function of changes in the inputs.
    The Jacobian of :math:`h(\nabla_{f,g} E = 0)` is a linear operator which, if
    it were to be instantiated as a matrix, would be of size
    :math:`(n+m) \times (n+m)`. When :math:`h` is the identity, that matrix is
    the Hessian of :math:`E`, is symmetric and negative-definite
    (:math:`E` is concave) and is structured as :math:`[A, B; B^T, D]`. More
    generally, for other functions :math:`h`, the Jacobian of these
    preconditioned
    first order conditions is no longer symmetric (except if ``a==b``), and
    has now a structure as :math:`[A, B; C, D]`. That system can
    be still inverted more generic solvers. By default, :math:`h = \epsilon
    \log`, as proposed in :cite:`cuturi:20a`.

    In both cases :math:`A` and :math:`D` are diagonal matrices, equal to the
    row and
    column marginals respectively, multiplied by the derivatives of :math:`h`
    evaluated at those marginals, corrected (if handling the unbalanced case)
    by the second derivative of the part of the objective that ties potentials
    to the marginals (terms in ``phi_star``). When :math:`h` is the identity,
    :math:`B` and :math:`B^T` are equal respectively to the OT matrix and its
    transpose, i.e. :math:`n \times m` and :math:`m \times n` matrices.
    When :math:`h` is not the identity, :math:`B` (resp. :math:`C`) is equal
    to the OT matrix (resp. its transpose), rescaled on the left by the
    application elementwise of :math:`h'` to the row (respectively column)
    marginal sum of the transport.

    Note that we take great care in not instantiating these transport
    matrices, to rely instead on calls to the ``app_transport`` method from the
    ``Geometry`` object ``geom`` (which will either use potentials or scalings,
    depending on ``lse_mode``)

    The Jacobian's diagonal + off-diagonal blocks structure allows to exploit
    Schur complements. Depending on the sizes involved, it is better to
    instantiate the Schur complement of the first or of the second diagonal
    block.

    In either case, the Schur complement is rank deficient, with a 0 eigenvalue
    for the vector of ones in the balanced case, which is why we add a ridge on
    that subspace to enforce solutions have zero sum.

    The Schur complement can also be rank deficient if two lines or columns of T
    are collinear. This will typically happen it two rows or columns of the cost
    or kernel matrix are numerically close. To avoid this, we add a more global
    ``ridge_identity * z`` regularizer to achieve better conditioning.

    These linear systems are solved using the user defined ``solver_fun``,
    which is set by default to ``cg``. When the system is symmetric (as detected
    by the corresponding flag ``symmetric``), ``cg`` is applied directly. When
    it is not, normal equations are used (i.e. the Schur complement is
    multiplied by its transpose before solving the system).

    Args:
      gr: 2-tuple, (vector of size ``n``, vector of size ``m``).
      ot_prob: the instantiation of the regularized transport problem.
      f: potential, w.r.t marginal a.
      g: potential, w.r.t marginal b.
      lse_mode: bool, log-sum-exp mode if True, kernel else.

    Returns:
      A tuple of two vectors, of the same size as ``gr``.
    """
    geom = ot_prob.geom
    marginal_a, marginal_b, app_transport = (
        ot_prob.get_transport_functions(lse_mode)
    )

    # elementwise vmap apply of derivative of precondition_fun. No vmapping
    # can be problematic here.
    if self.precondition_fun is None:
      precond_fun = lambda x: geom.epsilon * jnp.log(x)
    else:
      precond_fun = self.precondition_fun
    derivative = jax.vmap(jax.grad(precond_fun))

    n, m = geom.shape
    # pylint: disable=g-long-lambda
    vjp_fg = lambda z: app_transport(
        f, g, z * derivative(marginal_b(f, g)), axis=1
    ) / geom.epsilon
    vjp_gf = lambda z: app_transport(
        f, g, z * derivative(marginal_a(f, g)), axis=0
    ) / geom.epsilon

    if not self.symmetric:
      vjp_fgt = lambda z: app_transport(
          f, g, z, axis=0
      ) * derivative(marginal_b(f, g)) / geom.epsilon
      vjp_gft = lambda z: app_transport(
          f, g, z, axis=1
      ) * derivative(marginal_a(f, g)) / geom.epsilon

    diag_hess_a = (
        marginal_a(f, g) * derivative(marginal_a(f, g)) / geom.epsilon +
        unbalanced_functions.diag_jacobian_of_marginal_fit(
            ot_prob.a, f, ot_prob.tau_a, geom.epsilon, derivative
        )
    )
    diag_hess_b = (
        marginal_b(f, g) * derivative(marginal_b(f, g)) / geom.epsilon +
        unbalanced_functions.diag_jacobian_of_marginal_fit(
            ot_prob.b, g, ot_prob.tau_b, geom.epsilon, derivative
        )
    )

    n, m = geom.shape
    # Remove ridge on kernel space if problem is balanced.
    ridge_kernel = jnp.where(ot_prob.is_balanced, self.ridge_kernel, 0.0)

    # Forks on using Schur complement of either A or D, depending on size.
    if n > m:  #  if n is bigger, run m x m linear system.
      inv_vjp_ff = lambda z: z / diag_hess_a
      vjp_gg = lambda z: z * diag_hess_b
      schur_ = lambda z: vjp_gg(z) - vjp_gf(inv_vjp_ff(vjp_fg(z)))
      res = gr[1] - vjp_gf(inv_vjp_ff(gr[0]))

      if self.symmetric:
        schur = lambda z: (
            schur_(z) + ridge_kernel * jnp.sum(z) + self.ridge_identity * z
        )
      else:
        schur_t = lambda z: vjp_gg(z) - vjp_fgt(inv_vjp_ff(vjp_gft(z)))
        res = schur_t(res)
        schur = lambda z: (
            schur_t(schur_(z)) + ridge_kernel * jnp.sum(z) + self.ridge_identity
            * z
        )

      sch = self.solver_fun(schur, res)[0]
      vjp_gr_f = inv_vjp_ff(gr[0] - vjp_fg(sch))
      vjp_gr_g = sch
    else:
      vjp_ff = lambda z: z * diag_hess_a
      inv_vjp_gg = lambda z: z / diag_hess_b
      schur_ = lambda z: vjp_ff(z) - vjp_fg(inv_vjp_gg(vjp_gf(z)))
      res = gr[0] - vjp_fg(inv_vjp_gg(gr[1]))

      if self.symmetric:
        schur = lambda z: (
            schur_(z) + ridge_kernel * jnp.sum(z) + self.ridge_identity * z
        )
      else:
        schur_t = lambda z: vjp_ff(z) - vjp_gft(inv_vjp_gg(vjp_fgt(z)))
        res = schur_t(res)
        schur = lambda z: (
            schur_t(schur_(z)) + ridge_kernel * jnp.sum(z) + self.ridge_identity
            * z
        )

      sch = self.solver_fun(schur, res)[0]
      vjp_gr_g = inv_vjp_gg(gr[1] - vjp_gf(sch))
      vjp_gr_f = sch

    return jnp.concatenate((-vjp_gr_f, -vjp_gr_g))

  def first_order_conditions(
      self, prob, f: jnp.ndarray, g: jnp.ndarray, lse_mode: bool
  ):
    r"""Compute vector of first order conditions for the reg-OT problem.

    The output of this vector should be close to zero at optimality.
    Upon completion of the Sinkhorn forward pass, its norm (using the norm
    parameter defined using ``norm_error``) should be below the threshold
    parameter.

    This error will be itself assumed to be close to zero when using implicit
    differentiation.

    Args:
      prob: definition of the linear optimal transport problem.
      f: jnp.ndarray, first potential
      g: jnp.ndarray, second potential
      lse_mode: bool

    Returns:
      a jnp.ndarray of size (size of ``n + m``) quantifying deviation to
      optimality for variables ``f`` and ``g``.
    """
    geom = prob.geom
    marginal_a, marginal_b, _ = prob.get_transport_functions(lse_mode)
    grad_a = unbalanced_functions.grad_of_marginal_fit(
        prob.a, f, prob.tau_a, geom.epsilon
    )
    grad_b = unbalanced_functions.grad_of_marginal_fit(
        prob.b, g, prob.tau_b, geom.epsilon
    )
    if self.precondition_fun is None:
      precond_fun = lambda x: geom.epsilon * jnp.log(x)
    else:
      precond_fun = self.precondition_fun

    result_a = jnp.where(
        prob.a > 0,
        precond_fun(marginal_a(f, g)) - precond_fun(grad_a), 0.0
    )
    result_b = jnp.where(
        prob.b > 0,
        precond_fun(marginal_b(f, g)) - precond_fun(grad_b), 0.0
    )
    return jnp.concatenate((result_a, result_b))

  def gradient(
      self, prob: "linear_problem.LinearProblem", f: jnp.ndarray,
      g: jnp.ndarray, lse_mode: bool, gr: Tuple[jnp.ndarray, jnp.ndarray]
  ) -> "linear_problem.LinearProblem":
    """Apply vjp to recover gradient in reverse mode differentiation."""
    # Applies first part of vjp to gr: inverse part of implicit function theorem
    vjp_gr = self.solve(gr, prob, f, g, lse_mode)

    # Instantiates vjp of first order conditions of the objective, as a
    # function of geom, a and b parameters (against which we differentiate)
    foc_prob = lambda prob: self.first_order_conditions(prob, f, g, lse_mode)

    # Carries pullback onto original inputs, here geom, a and b.
    _, pull_prob = jax.vjp(foc_prob, prob)
    return pull_prob(vjp_gr)
